@@ -21,6 +21,21 @@ from firepit.sqlstorage import validate_name
 logger = logging.getLogger(__name__)
 
 
+MATCH_BIN = '''CREATE FUNCTION firepit_common.match_bin(pattern TEXT, value TEXT)
+RETURNS boolean AS $$
+    SELECT regexp_match(convert_from(decode(value, 'base64'), 'UTF8'), pattern) IS NOT NULL;
+$$ LANGUAGE SQL;'''
+
+LIKE_BIN = '''CREATE FUNCTION firepit_common.like_bin(pattern TEXT, value TEXT)
+RETURNS boolean AS $$
+    SELECT convert_from(decode(value, 'base64'), 'UTF8') LIKE pattern;
+$$ LANGUAGE SQL;'''
+
+COLUMNS_TABLE = ('CREATE UNLOGGED TABLE IF NOT EXISTS "__columns" '
+                 '(otype TEXT, path TEXT, shortname TEXT, dtype TEXT,'
+                 ' UNIQUE(otype, path));')
+
+
 def get_storage(url, session_id):
     dbname = url.path.lstrip('/')
     return PgStorage(dbname, url.geturl(), session_id)
@@ -74,7 +89,7 @@ class ListToTextIO:
         try:
             while n > len(self.buf):
                 obj = next(self.it)
-                vals = [ujson.dumps(val) if isinstance(val, list)
+                vals = [ujson.dumps(val, ensure_ascii=False) if isinstance(val, list)
                         else _text_encode(val) for val in obj]
                 self.buf += self.sep.join(vals) + '\n'
             result = self.buf[:n]
@@ -153,6 +168,7 @@ class PgStorage(SqlStorage):
     def __init__(self, dbname, url, session_id=None):
         super().__init__()
         self.placeholder = '%s'
+        self.dialect = 'postgresql'
         self.text_min = 'LEAST'
         self.text_max = 'GREATEST'
         self.ifnull = 'COALESCE'
@@ -194,23 +210,33 @@ class PgStorage(SqlStorage):
 
     def _create_firepit_common_schema(self):
         try:
-            stmt = ("SELECT  schema_name"
-                    " FROM INFORMATION_SCHEMA.SCHEMATA"
-                    " WHERE  schema_name = 'firepit_common'")
-            res = self._query(stmt).fetchone()
-            done = list(res.values())[0] if res else False
-            if not done:
+            stmt = ("SELECT routines.routine_name"
+                    " FROM information_schema.routines"
+                    " WHERE routines.specific_schema = 'firepit_common'")
+            res = self._query(stmt).fetchall()
+            if not res:
                 self._execute('CREATE SCHEMA IF NOT EXISTS "firepit_common";')
                 cursor = self._execute('BEGIN;')
                 self._execute('''CREATE FUNCTION firepit_common.match(pattern TEXT, value TEXT)
                                 RETURNS boolean AS $$
                                     SELECT regexp_match(value, pattern) IS NOT NULL;
                             $$ LANGUAGE SQL;''', cursor=cursor)
+                self._execute(MATCH_BIN, cursor=cursor)
+                self._execute(LIKE_BIN, cursor=cursor)
                 self._execute('''CREATE FUNCTION firepit_common.in_subnet(addr TEXT, net TEXT)
                                 RETURNS boolean AS $$
                                     SELECT addr::inet <<= net::inet;
                             $$ LANGUAGE SQL;''', cursor=cursor)
                 cursor.close()
+            elif len(res) < 4:
+                # Might need to add new functions
+                cursor = self._execute('BEGIN;')
+                funcs = [r['routine_name'] for r in res]
+                if 'match_bin' not in funcs:
+                    self._execute(MATCH_BIN, cursor=cursor)
+                if 'like_bin' not in funcs:
+                    self._execute(LIKE_BIN, cursor=cursor)
+                cursor.close()
         except psycopg2.errors.DuplicateFunction:
             self.connection.rollback()
 
@@ -231,6 +257,7 @@ class PgStorage(SqlStorage):
                     '(source_ref TEXT, target_ref TEXT, x_firepit_rank INTEGER,'
                     ' UNIQUE(source_ref, target_ref));')
             self._execute(stmt, cursor)
+            self._execute(COLUMNS_TABLE, cursor)
             self._set_meta(cursor, 'dbversion', DB_VERSION)
             self.connection.commit()
             cursor.close()
@@ -238,6 +265,12 @@ class PgStorage(SqlStorage):
             # We probably already created all these, so ignore this
             self.connection.rollback()
 
+    def _migrate(self, version, cursor):
+        if version == '2':
+            self._execute(COLUMNS_TABLE, cursor)
+            return True
+        return False
+
     def _get_writer(self, **kwargs):
         """Get a DB inserter object"""
         self.defer_index = kwargs.get('defer_index', self.defer_index)
@@ -469,7 +502,8 @@ class PgStorage(SqlStorage):
             if query_id and idx is not None:
                 query_values.append(obj[idx])
                 query_values.append(query_id)
-            values.extend([ujson.dumps(value) if isinstance(value, list) else value for value in obj])
+            values.extend([ujson.dumps(value, ensure_ascii=False)
+                           if isinstance(value, list) else value for value in obj])
         cursor.execute(stmt, values)
 
         if query_id and 'id' in colnames:
