@@ -4,6 +4,8 @@ import os
 import re
 import sqlite3
 
+from base64 import b64decode
+
 from firepit.exceptions import DuplicateTable
 from firepit.exceptions import InvalidAttr
 from firepit.exceptions import UnexpectedError
@@ -16,6 +18,11 @@ from firepit.sqlstorage import validate_name
 logger = logging.getLogger(__name__)
 
 
+COLUMNS_TABLE = ('CREATE TABLE IF NOT EXISTS "__columns" '
+                 '(otype TEXT, path TEXT, shortname TEXT, dtype TEXT,'
+                 ' UNIQUE(otype, path));')
+
+
 def get_storage(path):
     return SQLiteStorage(path)
 
@@ -32,7 +39,28 @@ def _in_subnet(value, net):
 
 def _match(pattern, value):
     """User-defined function to implement SQL MATCH/STIX MATCHES"""
-    return bool(re.match(pattern, value))
+    return (value is not None and
+            bool(re.search(pattern, value, re.DOTALL)))
+
+
+def _match_bin(pattern, value):
+    """User-defined function to implement SQL MATCH/STIX MATCHES for binaries"""
+    if value is not None:
+        val = b64decode(value).decode("utf-8")
+        return bool(re.search(pattern, val, re.DOTALL))
+    return False
+
+
+def _like_bin(pattern, value):
+    """User-defined function to implement SQL/STIX LIKE for binaries"""
+    if value is not None:
+        try:
+            exp = re.escape(pattern).replace('%', '.*').replace('_', '.')
+            val = b64decode(value).decode("utf-8")
+            return bool(re.search(exp, val, re.DOTALL))
+        except Exception as e:
+            logger.error('%s', e, exc_info=e)
+    return False
 
 
 class SQLiteStorage(SqlStorage):
@@ -49,6 +77,8 @@ class SQLiteStorage(SqlStorage):
 
         # Create function for SQL MATCH
         self.connection.create_function("match", 2, _match)
+        self.connection.create_function("match_bin", 2, _match_bin)
+        self.connection.create_function("like_bin", 2, _like_bin)
 
         cursor = self.connection.execute(
             "SELECT name FROM sqlite_master WHERE type='table' AND name='__queries'")
@@ -62,6 +92,12 @@ class SQLiteStorage(SqlStorage):
             self._initdb(cursor)
         cursor.close()
 
+    def _migrate(self, version, cursor):
+        if version == '2':
+            self._execute(COLUMNS_TABLE, cursor)
+            return True
+        return False
+
     def _get_writer(self, **kwargs):
         """Get a DB inserter object"""
         filedir = os.path.dirname(self.dbname)
