@@ -21,6 +21,7 @@ from firepit.props import parse_prop
 from firepit.props import primary_prop
 from firepit.query import Aggregation
 from firepit.query import Column
+from firepit.query import BinnedColumn
 from firepit.query import Filter
 from firepit.query import Group
 from firepit.query import Join
@@ -37,7 +38,7 @@ from firepit.stix21 import makeid
 from firepit.validate import validate_name
 from firepit.validate import validate_path
 
-DB_VERSION = "2"
+DB_VERSION = "2.1"
 
 logger = logging.getLogger(__name__)
 
@@ -48,7 +49,8 @@ def _transform(filename):
         if 'type' not in obj:
             obj['type'], _, _ = obj['id'].partition('--')
         if obj['type'] != 'identity':
-            for o in (raft.json_normalize(obj, flat_lists=False) for obj in raft.flatten(obj)):
+            for o in (raft.json_normalize(obj, flat_lists=False)
+                      for obj in raft.flatten(obj)):
                 yield o
         else:
             yield obj
@@ -76,6 +78,7 @@ class SqlStorage:
     def __init__(self):
         self.connection = None  # Python DB API connection object
         self.placeholder = '%s'  # Derived class can override this
+        self.dialect = None      # Derived class can override this
 
         # Functions to use for min/max text.  It can vary - sqlite3
         # uses MIN/MAX, postgresql uses LEAST/GREATEST
@@ -114,6 +117,10 @@ class SqlStorage:
                 '(source_ref TEXT, target_ref TEXT, x_firepit_rank,'
                 ' UNIQUE(source_ref, target_ref) ON CONFLICT IGNORE);')
         self._execute(stmt, cursor)
+        stmt = ('CREATE TABLE IF NOT EXISTS "__columns" '
+                '(otype TEXT, path TEXT, shortname TEXT, dtype TEXT,'
+                ' UNIQUE(otype, path));')
+        self._execute(stmt, cursor)
         self._set_meta(cursor, 'dbversion', DB_VERSION)
         self.connection.commit()
         cursor.close()
@@ -128,7 +135,13 @@ class SqlStorage:
         res = cursor.fetchone()
         dbversion = res['value'] if res else ""
         if dbversion != DB_VERSION:
-            raise DatabaseMismatch(dbversion, DB_VERSION)
+            if self._migrate(dbversion, cursor):
+                self._set_meta(cursor, 'dbversion', DB_VERSION)
+            else:
+                raise DatabaseMismatch(dbversion, DB_VERSION)
+
+    def _migrate(self, _version, _cursor):
+        return False
 
     def _set_meta(self, cursor, name, value):
         stmt = ('INSERT INTO "__metadata" (name, value)'
@@ -341,7 +354,7 @@ class SqlStorage:
             else:
                 action = 'NOTHING'
             stmt += f' ON CONFLICT (id) DO {action}'
-        values = tuple([ujson.dumps(value)
+        values = tuple([ujson.dumps(value, ensure_ascii=False)
                         if isinstance(value, list) else value for value in obj])
         logger.debug('_upsert: "%s", %s', stmt, values)
         cursor.execute(stmt, values)
@@ -703,7 +716,7 @@ class SqlStorage:
         return res
 
     def run_query(self, query):
-        query_text, query_values = query.render(self.placeholder)
+        query_text, query_values = query.render(self.placeholder, self.dialect)
         return self._query(query_text, query_values)
 
     def merge(self, viewname, input_views):
@@ -792,7 +805,7 @@ class SqlStorage:
             agg = Aggregation(aggs)
             query.aggs = agg
 
-        query_text, query_values = query.render('{}')
+        query_text, query_values = query.render('{}', self.dialect)
         formatted_values = [f"'{v}'" if isinstance(v, str) else v for v in query_values]
         stmt = query_text.format(*formatted_values)
         logger.debug('assign_query: %s', stmt)
@@ -958,14 +971,18 @@ class SqlStorage:
 
     def group(self, newname, viewname, by, aggs=None):
         """Create new view `newname` defined by grouping `viewname` by `by`"""
-        if isinstance(by, str):
+        if not isinstance(by, list):
             by = [by]
         columns = self.columns(viewname)
         group_colnames = []
         joined = set()
+        proj = []
         qry = Query(viewname)
         for col in by:
-            if col not in columns:
+            if isinstance(col, BinnedColumn):
+                proj.append(col)
+                group_colnames.append(col)
+            elif col not in columns:
                 joins, table, colname = self.path_joins(viewname, None, col)
                 group_colnames.append(Column(colname, table, col))
                 if table not in joined:
@@ -996,6 +1013,9 @@ class SqlStorage:
                 else:
                     tmp.append(agg)
             aggs = tmp
+        if proj:
+            qry.append(Projection(proj))
         qry.append(Group(group_colnames))
         qry.append(Aggregation(aggs))
+        qry.append(Order(group_colnames))
         self.assign_query(newname, qry)
